import { readFileSync } from 'fs'
import { expect } from 'chai'
import { sequelizer, SKIP, PASS } from './index.js'
import { nanoid } from 'nanoid'

const userSchema = JSON.parse(readFileSync('./fixtures/user.schema.json', 'utf-8'))
const addressSchema = JSON.parse(readFileSync('./fixtures/address.schema.json', 'utf-8'))

describe('json-sequelize', () => {
  describe('schema', () => {
    const [userModel, addressModel] = [sequelizer(userSchema), sequelizer(addressSchema)]

    it('should return an object', done => {
      expect(typeof userModel).to.equal('object')
      expect(typeof addressModel).to.equal('object')
      done()
    })

    it('should fail with unknown property value', done => {
      const schema = { properties: { foo: { type: 'bar' } } }
      expect(() => sequelizer(schema)).to.throw()
      done()
    })

    it('should contain correct number of fields', done => {
      expect(Object.keys(userModel).length).to.equal(Object.keys(userSchema.properties).length)
      expect(Object.keys(addressModel).length).to.equal(Object.keys(addressSchema.properties).length)
      done()
    })

    it('should have the correct type definitions', done => {
      expect(userModel.active.type.key).to.equal('BOOLEAN')
      expect(userModel.createdAt.type.key).to.equal('DATEONLY')
      expect(userModel.activeAt.type.key).to.equal('DATE')
      expect(userModel.username.type.key).to.equal('STRING')
      expect(userModel.email.type.key).to.equal('TEXT')
      expect(userModel.role.type.key).to.deep.equal('ENUM')
      expect(userModel.role.type.values).to.deep.equal(['admin', 'staff', 'guest'])
      expect(userModel.tags.type.key).to.equal('JSONB')
      expect(addressModel.streetName.type.key).to.equal('TEXT')
      expect(addressModel.lat.type.key).to.equal('REAL')
      expect(addressModel.lng.type.key).to.equal('REAL')
      done()
    })

    it('should have the proper unique fields', done => {
      expect(userModel.username.unique).to.exist.and.equal(true)
      expect(userModel.email.unique).to.exist.and.equal(true)
      done()
    })

    it('should have the proper allowNull fields', done => {
      expect(userModel.active.allowNull).to.exist.and.equal(false)
      expect(userModel.username.allowNull).to.exist.and.equal(false)
      expect(userModel.email.allowNull).to.exist.and.equal(false)
      done()
    })
  })

  describe('mixins', () => {
    describe('selector', () => {
      const schema = {
        properties: {
          id: { type: 'number', foo: 'bar' }
        }
      }

      it.skip('should not fail without a selector', done => {
        const mixin = {
          handler: ({ DataTypes }) => ({ type: DataTypes.STRING })
        }
        // expect(() => sequelizer(schema, mixin)).to.throw('Mixin is missing a "selector"')
        done()
      })

      it('should accept a selector function', done => {
        const mixin = {
          selector: ({ key }) => key === 'id',
          handler: ({ DataTypes }) => ({ type: DataTypes.STRING })
        }
        const model = sequelizer(schema, mixin)
        expect(model.id.type.key).to.equal('STRING')
        done()
      })

      it('should accept a selector string matching a properties key', done => {
        const schema = {
          properties: {
            foo: { type: 'number' },
            bar: { type: 'number' }
          }
        }
        const mixin = {
          selector: 'foo',
          handler: ({ DataTypes }) => ({ type: DataTypes.TEXT })
        }
        const model = sequelizer(schema, mixin)
        expect(model.foo.type.key).to.equal('TEXT')
        expect(model.bar.type.key).to.equal('REAL')
        done()
      })

      it('should accept a selector object with a matching property object', done => {
        const mixin1 = {
          selector: { type: 'number' },
          handler: ({ DataTypes }) => ({ type: DataTypes.STRING })
        }
        expect(sequelizer(schema, mixin1).id.type.key).to.equal('STRING')

        const mixin2 = {
          selector: { foo: 'bar' },
          handler: ({ DataTypes }) => ({ type: DataTypes.STRING })
        }
        expect(sequelizer(schema, mixin2).id.type.key).to.equal('STRING')
        done()
      })

      it('should accept a selector object with a property key', done => {
        const mixin = {
          selector: { key: 'id' },
          handler: ({ DataTypes }) => ({ type: DataTypes.STRING })
        }
        const model = sequelizer(schema, mixin)
        expect(model.id.type.key).to.equal('STRING')
        done()
      })
    })

    describe('handler', () => {
      it('should fail without a handler', done => {
        const mixin = {
          selector: { key: 'id' }
        }
        // @ts-ignore
        expect(() => sequelizer({ properties: { id: {} } }, mixin)).to.throw('Mixin is missing "handler" function')
        done()
      })

      it('should fail with a bad handler', done => {
        const mixin = {
          selector: { key: 'id' },
          handler: () => {}
        }
        expect(() => sequelizer({ properties: { id: {} } }, mixin)).to.throw(
          '"handler" function must return a response'
        )
        done()
      })

      it('should accept a handler function that sets custom model defintions', done => {
        const mixin = {
          selector: { key: 'id' },
          handler: ({ DataTypes }) => ({
            type: DataTypes.STRING,
            defaultValue: nanoid,
            primaryKey: true
          })
        }
        const model = sequelizer({ properties: { id: {} } }, mixin)
        expect(model.id.type.key).to.equal('STRING')
        expect(model.id.primaryKey).to.equal(true)
        expect(model.id.defaultValue).to.equal(nanoid)
        done()
      })

      it('should accept a handler function with plain string types', done => {
        const mixin = {
          selector: { key: 'id' },
          handler: () => ({
            type: 'STRING',
            primaryKey: true
          })
        }
        const model = sequelizer({ properties: { id: {} } }, mixin)
        expect(model.id.type.key).to.equal('STRING')
        expect(model.id.primaryKey).to.equal(true)
        done()
      })

      it('should accept a handler function that removes schema properties from model definitions', done => {
        const schema = {
          computed: ['beep', 'boop'],
          properties: {
            id: { type: 'string' },
            beep: {},
            boop: {}
          }
        }
        const mixin = {
          handler: ({ schema, key, SKIP }) => {
            return schema.computed.includes(key) && SKIP
          }
        }
        const model = sequelizer(schema, mixin)
        expect(Object.keys(model).length).to.equal(1)
        done()
      })
    })

    describe('multiple', () => {
      it('should accept multiple mixins', done => {
        const schema = {
          computed: ['beep', 'boop'],
          properties: {
            id: {},
            beep: {},
            boop: {}
          }
        }
        const mixin1 = {
          selector: { key: 'id' },
          handler: () => ({ type: 'STRING', primaryKey: true })
        }
        const mixin2 = {
          handler: ({ schema, key, SKIP }) => schema.computed.includes(key) && SKIP
        }
        const model = sequelizer(schema, [mixin1, mixin2])
        expect(model.id.type.key).to.equal('STRING')
        expect(model.id.primaryKey).to.equal(true)
        expect(Object.keys(model).length).to.equal(1)
        done()
      })
    })
  })
})
